# -*- coding: utf-8 -*-
# Copyright 2010 Canonical Ltd.  This software is licensed under the
# GNU Lesser General Public License version 3 (see the file LICENSE).

import httplib2
import simplejson
import urllib
from functools import wraps

class APIError(Exception):
    def __init__(self, msg, body=None):
        self.msg = msg
        self.body = body
    def __str__(self):
        return self.msg

def returns_json(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        response, body = func(*args, **kwargs)
        return simplejson.loads(body)
    return wrapper


def returns(cls, none_allowed=False):
    """The response data will be deserialized into an instance of 'cls'.

    The provided class should be a descendant of PistonResponseObject,
    or some other class that provides a 'from_response' method.

    'none_allowed', defaulting to False, specifies whether or not None
    is a valid response. If True then the api can return None instead
    of a PistonResponseObject.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            response, body = func(self, *args, **kwargs)
            return cls.from_response(response, body, none_allowed)
        return wrapper
    return decorator

def returns_list_of(cls):
    """The response data will be deserialized into a list of 'cls'.

    The provided class should be a descendant of PistonResponseObject,
    or some other class that provides a 'from_response' method.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            response, body = func(self, *args, **kwargs)
            data = simplejson.loads(body)
            items = []
            for datum in data:
                items.append(cls.from_dict(datum))
            return items
        return wrapper
    return decorator


class PistonResponseObject(object):
    """Base class for objects that are returned from api calls.
    """
    @classmethod
    def from_response(cls, response, body, none_allowed=False):
        data = simplejson.loads(body)
        if none_allowed and data is None:
            return data
        obj = cls.from_dict(data)
        return obj

    @classmethod
    def from_dict(cls, data):
        obj = cls()
        for key, value in data.items():
            setattr(obj, key, value)
        return obj

class PistonSerializable(object):
    """Base class for objects that want to be used as api call arguments.
    """
    _atts = ()

    def __init__(self, **kwargs):
        for (key, value) in kwargs.items():
            setattr(self, key, value)

    def _as_serializable(self):
        data = {}
        for att in self._atts:
            if not hasattr(self, att):
                raise ValueError("Attempted to serialize attribute '%s'"
                    % att)
            data[att] = getattr(self, att)
        return data


class PistonAPI(object):
    """This class provides methods to make http requests slightly easier.

    It's a small wrapper around httplib2 to allow for a bit of state to
    be stored (like the service root) so that you don't need to repeat
    yourself as much.
    It's not intended to be used directly.  Children classes should implement
    methods that actually call out to the api methods.

    When you define your API's methods you'll
    want to just call out the _get, _post methods provided by this class.
    """

    default_service_root = ''

    default_content_type = 'application/json'

    def __init__(self, service_root=None, cachedir=None, auth=None):
        if service_root is None:
            service_root = self.default_service_root
        self._service_root = service_root
        self._cachedir = cachedir
        self._auth = auth

    def _get(self, path, args=None):
        """Perform an HTTP GET request.

        The provided 'path' is appended to this resource's '_service_root'
        attribute to obtain the absolute URL that will be requested.

        If provided, 'args' should be a dict specifying additional GET
        arguments that will be encoded on to the end of the url.
        """
        if args is not None:
            if '?' in path:
                path += '&'
            else:
                path += '?'
            path += urllib.urlencode(args)
        return self._request(path, method='GET')

    def _post(self, path, data=None, content_type=None):
        """Perform an HTTP POST request.

        The provided path is appended to this api's '_service_root' attribute
        to obtain the absolute URL that will be requested.  Data should be:
         - A string, in which case it will be used directly as the request's
           body, or
         - A list, dict, int, bool or PistonSerializable (something with
           an '_as_serializable' method) or even None, in
           which case it will be encoded into 'content_type' using the right
           serializer from serializers.

        If content_type is None, self.default_content_type will be used.
        """
        body = data
        if content_type is None:
            content_type = self.default_content_type
        if not isinstance(data, basestring):
            # Import here to avoid a circular import
            from piston_mini_client.serializers import get_serializer
            serializer = get_serializer(content_type)(data)
            body = serializer.serialize()
        headers = {'Content-type': content_type}
        return self._request(path, method='POST', body=body, headers=headers)

    def _request(self, path, method, body='', headers=None):
        """Perform an HTTP request.

        You probably want to call one of the _get, _post methods instead.
        """
        http = httplib2.Http(cache=self._cachedir)
        if headers is None:
            headers = {}
        url = self._path2url(path)
        if self._auth:
            self._auth.sign_request(url, method, body, headers)
        try:
            response, body = http.request(url, method=method, body=body,
                headers=headers)
        except AttributeError, e:
            # Special case out httplib2's way of telling us unable to connect
            if e.args[0] == "'NoneType' object has no attribute 'makefile'":
                raise APIError('Unable to connect to %s' % self._service_root)
            else:
                raise
        if response['status'] not in ['200', '201', '304']:
            raise APIError('%s: %s' % (response['status'],response), body)
        return response, body

    def _path2url(self, path):
        return self._service_root.rstrip('/') + '/' + path.lstrip('/')

