from mock import patch
from unittest import TestCase
from wsgi_intercept import add_wsgi_intercept, remove_wsgi_intercept
from wsgi_intercept.httplib2_intercept import install, uninstall

from piston_mini_client import (PistonAPI, returns_json, returns,
    returns_list_of, PistonResponseObject, PistonSerializable)
from piston_mini_client.auth import BasicAuthorizer


class PistonAPITestCase(TestCase):
    # XXX achuni 2010-11-30: We need to test different return codes
    class CoffeeAPI(PistonAPI):
        default_service_root = 'http://localhost:12345'
        def brew(self):
            self._get('/brew')

    def callback(self, environ, start_response):
        self.called = {
            'body': environ['wsgi.input'].read(),
            'path': environ['PATH_INFO'],
            'auth': environ.get('HTTP_AUTHORIZATION'),
            'content_type': environ.get('CONTENT_TYPE'),
            'query': environ.get('QUERY_STRING'),
        }
        return 'hello world'

    def setUp(self):
        install()
        add_wsgi_intercept('localhost', 12345, lambda: self.callback)

    def tearDown(self):
        remove_wsgi_intercept('localhost', 12345)
        uninstall()
        
    def test_request(self):
        api = self.CoffeeAPI()
        api._request('/foo', 'POST', body='foo=bar')
        self.assertEqual(self.called['body'], 'foo=bar')
        self.assertEqual(self.called['path'], '/foo')

    def test_auth_request(self):
        api = self.CoffeeAPI(auth=BasicAuthorizer(username='foo',
            password='bar'))
        api._request('/fee', 'GET')
        self.assertEqual(self.called['auth'], 'Basic Zm9vOmJhcg==')

    def test_post_no_content_type(self):
        api = self.CoffeeAPI()
        api._post('/serve', data={'foo': 'bar'})
        self.assertEqual(self.called['content_type'], 'application/json')

    def test_post_piston_serializable(self):
        class MyCoffeeRequest(PistonSerializable):
            _atts = ('strength',)
        api = self.CoffeeAPI()
        api._post('/serve', data=MyCoffeeRequest(strength='mild'))
        self.assertEqual(self.called['content_type'], 'application/json')

    def test_post_explicit_content_type(self):
        api = self.CoffeeAPI()
        api._post('/serve', data={'foo': 'bar'},
            content_type='application/x-www-form-urlencoded')
        self.assertEqual(self.called['content_type'],
            'application/x-www-form-urlencoded')

    def test_get_no_args(self):
        api = self.CoffeeAPI()
        api._get('/stew')
        self.assertEqual('/stew', self.called['path'])
        self.assertEqual('', self.called['query'])

    def test_get_with_args(self):
        api = self.CoffeeAPI()
        api._get('/stew', args={'foo': 'bar'})
        self.assertEqual('/stew', self.called['path'])
        self.assertEqual('foo=bar', self.called['query'])

    @patch('httplib2.Http.request')
    def test_valid_status_codes_dont_raise_exception(self, mock_request):
        for status in ['200', '201', '304']:
            expected_response = {'status': status}
            expected_body = '"hello world!"'
            mock_request.return_value = (expected_response, expected_body)
            api = self.CoffeeAPI()
            response, body = api._get('/simmer')
            self.assertEqual(expected_response, response)

    def test_get_with_extra_args(self):
        api = self.CoffeeAPI()
        api._get('/stew?zot=ping', args={'foo': 'bar'})
        self.assertEqual('/stew', self.called['path'])
        self.assertEqual('zot=ping&foo=bar', self.called['query'])

    def test_path2url_with_no_ending_slash(self):
        resource = PistonAPI()
        resource._service_root = 'http://example.com/api'
        expected = 'http://example.com/api/frobble'
        self.assertEqual(expected, resource._path2url('frobble'))

    def test_path2url_with_ending_slash(self):
        resource = PistonAPI()
        resource._service_root = 'http://example.com/api/'
        expected = 'http://example.com/api/frobble'
        self.assertEqual(expected, resource._path2url('frobble'))


class PistonResponseObjectTestCase(TestCase):
    def test_from_response(self):
        obj = PistonResponseObject.from_response({}, '{"foo": "bar"}')
        self.assertEqual('bar', obj.foo)

    def test_from_dict(self):
        obj = PistonResponseObject.from_dict({"foo": "bar"})
        self.assertEqual('bar', obj.foo)


class ReturnsJSONTestCase(TestCase):
    def test_returns_json(self):
        class MyAPI(PistonAPI):
            default_service_root = 'foo'
            @returns_json
            def func(self):
                return {}, '{"foo": "bar", "baz": 42}'

        result = MyAPI().func()
        self.assertEqual({"foo": "bar", "baz": 42}, result)


class ReturnsTestCase(TestCase):
    def test_returns(self):
        class MyAPI(PistonAPI):
            default_service_root = 'foo'
            @returns(PistonResponseObject)
            def func(self):
                return {}, '{"foo": "bar", "baz": 42}'

        result = MyAPI().func()
        self.assertTrue(isinstance(result, PistonResponseObject))

    def test_returns_none_allowed(self):
        class MyAPI(PistonAPI):
            default_service_root = 'foo'
            @returns(PistonResponseObject, none_allowed=True)
            def func(self):
                return {}, 'null'

        result = MyAPI().func()
        self.assertEqual(result, None)

    def test_returns_none_allowed_normal_response(self):
        class MyAPI(PistonAPI):
            default_service_root = 'foo'
            @returns(PistonResponseObject, none_allowed=True)
            def func(self):
                return {}, '{"foo": "bar", "baz": 42}'

        result = MyAPI().func()
        self.assertTrue(isinstance(result, PistonResponseObject))


class ReturnsListOfTestCase(TestCase):
    def test_returns(self):
        class MyAPI(PistonAPI):
            default_service_root = 'foo'
            @returns_list_of(PistonResponseObject)
            def func(self):
                return {}, '[{"foo": "bar"}, {"baz": 42}]'

        result = MyAPI().func()
        self.assertEqual(2, len(result))
        self.assertEqual('bar', result[0].foo)
        self.assertEqual(42, result[1].baz)


class PistonSerializableTestCase(TestCase):
    class MySerializable(PistonSerializable):
        _atts = ('foo',)

    def test_init_with_extra_variables(self):
        obj = self.MySerializable(foo='bar', baz=42)
        self.assertEqual('bar', obj.foo)
        self.assertEqual(42, obj.baz)

    def test_init_with_missing_variables(self):
        obj = self.MySerializable()
        self.assertFalse(hasattr(obj, 'foo'))

    def test_missing_required_arguments(self):
        obj = self.MySerializable()
        self.assertRaises(ValueError, obj._as_serializable)

    def test_can_assign_required_arguments_after_init(self):
        obj = self.MySerializable()
        obj.foo = 'bar'
        self.assertEqual({'foo': 'bar'}, obj._as_serializable())

    def test_extra_args_arent_serialized(self):
        obj = self.MySerializable(foo='bar', baz=42)
        self.assertEqual({'foo': 'bar'}, obj._as_serializable())

